"""
The `cli` interface to the `PyZEAL` package.

Right now the command line interface exposes functionality related to

- viewing and changing of options
- viewing and (un-)installing of plugins
- viewing basic information like help and package version
"""
